(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.jsonh = {})));
}(this, (function (exports) { 'use strict';

  function _typeof(obj) {
    if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
      _typeof = function (obj) {
        return typeof obj;
      };
    } else {
      _typeof = function (obj) {
        return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
      };
    }

    return _typeof(obj);
  }

  function _slicedToArray(arr, i) {
    return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
  }

  function _arrayWithHoles(arr) {
    if (Array.isArray(arr)) return arr;
  }

  function _iterableToArrayLimit(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"] != null) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  function _nonIterableRest() {
    throw new TypeError("Invalid attempt to destructure non-iterable instance");
  }

  var format_1 = format;

  function format(target) {
    var result = '';
    var indent = 0;
    appendValue(target, '', true);
    return result;

    function appendValue(value) {
      var delimiter = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '\n';
      var root = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

      var type = _typeof(value);

      if (type === 'number' || type === 'boolean' || type === 'undefined' || value === null) {
        append(JSON.stringify(value) + delimiter);
      } else if (type === 'string') {
        append(util.quoteString(value, true) + delimiter);
      } else if (type === 'object' && Array.isArray(value)) {
        if (value.length === 0) {
          append('[]' + delimiter);
        } else {
          var nextDelimiter = '\n';
          append('[' + nextDelimiter);
          indent += 2;
          var _iteratorNormalCompletion = true;
          var _didIteratorError = false;
          var _iteratorError = undefined;

          try {
            for (var _iterator = value[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
              var element = _step.value;
              appendIndent();
              appendValue(element, nextDelimiter);
            }
          } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion && _iterator.return != null) {
                _iterator.return();
              }
            } finally {
              if (_didIteratorError) {
                throw _iteratorError;
              }
            }
          }

          indent -= 2;
          appendIndent();
          append(']' + delimiter);
        }
      } else if (type === 'object') {
        var entries = Object.entries(value);
        entries = entries.filter(function (_ref) {
          var _ref2 = _slicedToArray(_ref, 2),
              k = _ref2[0],
              v = _ref2[1];

          return v !== undefined;
        });

        if (entries.length === 0) {
          append('{}' + delimiter);
        } else {
          var _nextDelimiter = '\n';
          var doBracket = !(entries.length === 1 || root === true);

          if (doBracket) {
            append('{' + _nextDelimiter);
            indent += 2;
          }

          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            for (var _iterator2 = entries[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              var _step2$value = _slicedToArray(_step2.value, 2),
                  k = _step2$value[0],
                  v = _step2$value[1];

              if (doBracket) {
                appendIndent();
              }

              append(util.quoteString(k) + ': ');
              appendValue(v, _nextDelimiter);
            }
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return != null) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }

          if (doBracket) {
            indent -= 2;
            appendIndent();
            append('}' + delimiter);
          }
        }
      }
    }

    function appendIndent() {
      append(' '.repeat(indent));
    }

    function append(str) {
      result += str;
    }
  }

  var util = {
    quoteString: function quoteString(value) {
      var asValue = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
      var isQuote = false;
      var quotes = {
        '\'': 0.1,
        '"': 0.2
      };
      var replacements = {
        '\'': '\\\'',
        '"': '\\"',
        '\\': '\\\\',
        '\b': '\\b',
        '\f': '\\f',
        '\n': '\\n',
        '\r': '\\r',
        '\t': '\\t',
        '\v': '\\v',
        '\0': '\\0',
        "\u2028": "\\u2028",
        "\u2029": "\\u2029"
      };
      var product = '';
      var _iteratorNormalCompletion3 = true;
      var _didIteratorError3 = false;
      var _iteratorError3 = undefined;

      try {
        for (var _iterator3 = value[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
          var c = _step3.value;

          switch (c) {
            case '\'':
            case '"':
              quotes[c]++;
              product += c;
              isQuote = true;
              continue;
          }

          if (util.isSpaceSeparator(c) || util.isSpecial(c)) {
            isQuote = true;
          }

          if (replacements[c]) {
            product += replacements[c];
            isQuote = true;
            continue;
          }

          if (c < ' ') {
            var hexString = c.charCodeAt(0).toString(16);
            product += '\\x' + ('00' + hexString).substring(hexString.length);
            isQuote = true;
            continue;
          }

          product += c;
        }
      } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion3 && _iterator3.return != null) {
            _iterator3.return();
          }
        } finally {
          if (_didIteratorError3) {
            throw _iteratorError3;
          }
        }
      }

      if (!isQuote && asValue) {
        if (value !== util.translate(value)) {
          isQuote = true;
        }
      }

      if (!isQuote) {
        return value;
      }

      var quoteChar = Object.keys(quotes).reduce(function (a, b) {
        return quotes[a] < quotes[b] ? a : b;
      });
      product = product.replace(new RegExp(quoteChar, 'g'), replacements[quoteChar]);
      return quoteChar + product + quoteChar;
    },
    translate: function translate(rawString) {
      switch (rawString) {
        case 'null':
          return null;

        case 'true':
          return true;

        case 'false':
          return false;
      }

      if (!isNaN(rawString)) {
        return rawString - 0;
      }

      return rawString;
    },
    isSpaceSeparator: function isSpaceSeparator(c) {
      switch (c) {
        case '\t':
        case '\v':
        case '\f':
        case ' ':
        case "\xA0":
        case "\uFEFF":
        case '\n':
        case '\r':
        case "\u2028":
        case "\u2029":
          return true;
      }

      return /[\u1680\u2000-\u200A\u202F\u205F\u3000]/.test(c);
    },
    isSpecial: function isSpecial(c) {
      switch (c) {
        case ',':
        case ';':
        case '[':
        case ']':
        case '{':
        case '}':
        case ':':
          return true;
      }

      return false;
    }
  };

  var source;
  var parseState;
  var pos;
  var line;
  var column;
  var token;
  var root;

  var parse = function parse(text, reviver) {
    source = String(text);
    parseState = 'start';
    pos = 0;
    line = 1;
    column = 0;
    token = undefined;
    root = undefined;

    do {
      token = lex(); // This code is unreachable.
      // if (!parseStates[parseState]) {
      //     throw invalidParseState()
      // }

      parseStates[parseState]();
    } while (token.type !== 'eof');

    if (typeof reviver === 'function') {
      return internalize({
        '': root
      }, '', reviver);
    }

    return root;
  };

  function internalize(holder, name, reviver) {
    var value = holder[name];

    if (value != null && _typeof(value) === 'object') {
      for (var _key in value) {
        if (Object.hasOwnProperty.call(value, _key)) {
          var replacement = internalize(value, _key, reviver);

          if (replacement === undefined) {
            delete value[_key];
          } else {
            value[_key] = replacement;
          }
        }
      }
    }

    return reviver.call(holder, name, value);
  }

  var lexState;
  var buffer;
  var escaped;
  var quote;
  var c;
  var level;

  function lex() {
    lexState = 'default';
    buffer = '';
    level = 0;
    escaped = false;
    quote = 0;

    for (;;) {
      c = peek(); // This code is unreachable.
      // if (!lexStates[lexState]) {
      //     throw invalidLexState(lexState)
      // }

      var _token = lexStates[lexState]();

      if (_token) {
        return _token;
      }
    }
  }

  function peek() {
    if (source[pos]) {
      return String.fromCodePoint(source.codePointAt(pos));
    }
  } // pop peek result and modify c l p


  function read() {
    var c = peek();

    if (c === '\n') {
      line++;
      column = 0;
    } else if (c) {
      column += c.length;
    } else {
      column++;
    }

    if (c) {
      pos += c.length;
    }

    return c;
  }

  var lexStates = {
    default: function _default() {
      if (c === undefined) {
        read();
        return newToken('eof');
      }

      if (util$1.isSpaceSeparator(c)) {
        read();
        return;
      }

      lexState = 'value';
    },
    rawString: function rawString() {
      if (c === '\\') {
        read();
        buffer += escape();
        escaped = true;
        return;
      }

      if (util$1.isSpaceSeparator(c) || util$1.isSpecial(c) || c === undefined) {
        if (c === '[' || c === '{') {
          throw invalidChar(read());
        }

        if (buffer === '') {
          throw invalidChar(read());
        }

        lexState = 'stringBuilt';
        return;
      }

      buffer += read();
    },
    string: function string() {
      switch (c) {
        case '\\':
          read();
          buffer += escape();
          return;

        case '"':
        case '\'':
          if (quote === 1 && c === '\'' || quote === 2 && c === '"') {
            read();
            lexState = 'stringBuilt';
            var next = peek();

            if (!(util$1.isSpaceSeparator(next) || util$1.isSpecial(next) || next === undefined)) {
              throw invalidChar(read());
            }

            if (next === '[' || next === '{') {
              throw invalidChar(read());
            }

            return;
          }

          buffer += read();
          return;

        case '\n':
        case '\r':
          throw invalidChar(read());
        //      case '\u2028':
        //      case '\u2029':
        // Accept

        case undefined:
          throw invalidChar(read());
      }

      buffer += read();
    },
    delimiter: function delimiter() {
      if (c === ';') {
        buffer += read();
        level += 2;
        return;
      } // will return token seeing , or other


      if (c === ',') {
        if (level % 2 === 0) {
          throw invalidChar(read());
        }

        buffer += read();
        level += 1;
      }

      var token = newToken('delimiter', buffer);
      token.level = level;
      return token;
    },
    stringBuilt: function stringBuilt() {
      if (util$1.isSpaceSeparator(c)) {
        read();
        return;
      }

      if (c === ':') {
        read();
        return newToken('key', buffer);
      }

      if (quote > 0 || escaped === true) {
        return newToken('string', buffer);
      }

      return newToken('rawString', buffer);
    },
    closePunctuator: function closePunctuator() {
      if (util$1.isSpaceSeparator(c)) {
        return newToken('punctuator', buffer);
      }

      if (c === '}' || c === ']' || c === ',' || c === ';' || c === undefined) {
        return newToken('punctuator', buffer);
      }

      throw invalidChar(c);
    },
    value: function value() {
      switch (c) {
        case '{':
        case '[':
          return newToken('punctuator', read());

        case '}':
        case ']':
          buffer = read();
          lexState = 'closePunctuator';
          return;

        case ',':
        case ';':
          if (c === ',') {
            level = 0;
          } else {
            level = 1;
          }

          lexState = 'delimiter';
          buffer = read();
          return;

        case '"':
        case '\'':
          if (read() === '"') {
            quote = 2;
          } else {
            quote = 1;
          }

          buffer = '';
          lexState = 'string';
          return;

        case undefined:
          throw invalidChar(read());
      }

      lexState = 'rawString';
    }
  }; //

  function newToken(type, value) {
    return {
      type: type,
      value: value,
      line: line,
      column: column
    };
  }

  function escape() {
    var c = peek();

    switch (c) {
      case 'b':
        read();
        return '\b';

      case 'f':
        read();
        return '\f';

      case 'n':
        read();
        return '\n';

      case 'r':
        read();
        return '\r';

      case 't':
        read();
        return '\t';

      case 'v':
        read();
        return '\v';

      case '0':
        read();

        if (util$1.isDigit(peek())) {
          throw invalidChar(read());
        }

        return '\0';

      case 'x':
        read();
        return hexEscape();

      case 'u':
        read();
        return unicodeEscape();

      case '\n':
      case "\u2028":
      case "\u2029":
        read();
        return '';

      case '\r':
        read();

        if (peek() === '\n') {
          read();
        }

        return '';

      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        throw invalidChar(read());

      case undefined:
        throw invalidChar(read());
    }

    return read();
  }

  function hexEscape() {
    var buffer = '';
    var c = peek();

    if (!util$1.isHexDigit(c)) {
      throw invalidChar(read());
    }

    buffer += read();
    c = peek();

    if (!util$1.isHexDigit(c)) {
      throw invalidChar(read());
    }

    buffer += read();
    return String.fromCodePoint(parseInt(buffer, 16));
  }

  function unicodeEscape() {
    var buffer = '';
    var count = 4;

    while (count > 0) {
      count--;

      var _c = peek();

      if (!util$1.isHexDigit(_c)) {
        throw invalidChar(read());
      }

      buffer += read();
    }

    return String.fromCodePoint(parseInt(buffer, 16));
  }

  var key;
  var stack;
  var parseStates = {
    start: function start() {
      stack = [];
      key = undefined;
      parseStates['beforeValue']();
    },
    beforeValue: function beforeValue() {
      if (token.type === 'eof') {
        throw invalidEOF();
      }

      if (token.type === 'delimiter') {
        throw invalidToken();
      }

      if (token.type === 'punctuator') {
        if (token.value === ']') {
          throw invalidToken();
        }

        if (token.value === '}') {
          throw invalidToken();
        }
      }

      push();
    },
    beforeArrValue: function beforeArrValue() {
      if (token.type === 'eof') {
        throw invalidEOF();
      }

      if (token.type === 'delimiter') {
        throw invalidToken();
      }

      if (token.type === 'punctuator') {
        if (token.value === ']') {
          return popUntilArr();
        }

        if (token.value === '}') {
          throw invalidToken();
        }
      }

      push();
    },
    beforeKey: function beforeKey() {
      if (token.type === 'eof') {
        throw invalidEOF();
      }

      if (token.type === 'delimiter') {
        throw invalidToken();
      }

      if (token.type === 'punctuator') {
        if (token.value === '}') {
          return popUntilObj();
        } else {
          throw invalidToken();
        }
      }

      push();
    },
    afterValue: function afterValue() {
      if (token.type === 'eof') {
        return popUntilEof();
      }

      if (token.type === 'punctuator') {
        if (token.value === ']') {
          return popUntilArr();
        }

        if (token.value === '}') {
          return popUntilObj();
        }
      }

      if (token.type === 'delimiter') {
        popUntilLevel(token.level);
        parseState = 'beforeKey';
        return;
      }

      popUntilLevel(0);
      push();
    }
  };

  function translate(rawString) {
    switch (rawString) {
      case 'null':
        return null;

      case 'true':
        return true;

      case 'false':
        return false;
    }

    if (!isNaN(rawString)) {
      return rawString - 0;
    }

    return rawString;
  }

  function push() {
    var value;
    var info;

    switch (token.type) {
      case 'punctuator':
        switch (token.value) {
          case '{':
            value = {};
            info = {
              type: 'obj',
              value: value,
              limitLevel: 0
            };
            break;

          case '[':
            value = [];
            info = {
              type: 'arr',
              value: value,
              limitLevel: 0
            };
            break;
        }

        break;

      case 'rawString':
      case 'string':
        if (token.type === 'rawString') {
          value = translate(token.value);
        } else {
          value = token.value;
        }

        info = {
          type: 'value',
          value: value
        };
        break;

      case 'key':
        value = {};
        info = {
          type: 'obj_h',
          value: value,
          limitLevel: 0
        };
        break;
    }

    var parent = stack[stack.length - 1];

    if (root === undefined) {
      root = value;
    } else if (info) {
      if (parent.type === 'obj' || parent.type === 'obj_h') {
        if (info.type === 'obj_h' && key === undefined) {
          key = token.value;
          parseState = 'beforeValue';
          return;
        }

        parent.value[key] = info.value;
        key = undefined;
      } else if (parent.type === 'arr') {
        parent.value.push(info.value);
      }

      if (info.type === 'obj_h') {
        info.limitLevel = parent.level || parent.limitLevel || 0;
      }
    }

    if (info.type === 'value') {
      parseState = 'afterValue';
      return;
    }

    if (info.type === 'obj') {
      parseState = 'beforeKey';
    } else if (info.type === 'arr') {
      parseState = 'beforeArrValue';
    } else if (info.type === 'obj_h') {
      key = token.value;
      parseState = 'beforeValue';
    }

    stack.push(info);
  }

  function popUntilLevel(level) {
    while (true) {
      if (stack.length === 0) {
        throw invalidToken();
      }

      var current = stack[stack.length - 1];

      if (current.type === 'arr') {
        if (level > 0) {
          throw invalidToken();
        }

        return;
      }

      if (current.level != null) {
        if (current.level === level) {
          return;
        }

        if (current.level < level) {
          throw invalidToken();
        }

        stack.pop();
        continue;
      }

      if (level > current.limitLevel) {
        current.level = level;
        return;
      }

      if (level === 0) {
        if (current.type === 'obj' || stack.length === 1) {
          current.level = 0;
          return;
        }
      }

      stack.pop();
    }
  }

  function popUntilEof() {
    while (stack.length > 0) {
      var current = stack.pop();

      if (current.type === 'arr') {
        throw invalidToken();
      }

      if (current.type === 'obj') {
        throw invalidToken();
      }
    }
  }

  function popUntilObj() {
    do {
      if (stack.length === 0) {
        throw invalidToken();
      }

      var current = stack.pop();

      if (current.type === 'arr') {
        throw invalidToken();
      }
    } while (current.type !== 'obj');

    parseState = 'afterValue';
  }

  function popUntilArr() {
    do {
      if (stack.length === 0) {
        throw invalidToken();
      }

      var current = stack.pop();

      if (current.type === 'obj') {
        throw invalidToken();
      }
    } while (current.type !== 'arr');

    parseState = 'afterValue';
  } // This code is unreachable.
  // function invalidParseState () {
  //     return new Error(`JSON5: invalid parse state '${parseState}'`)
  // }
  // This code is unreachable.
  // function invalidLexState (state) {
  //     return new Error(`JSON5: invalid lex state '${state}'`)
  // }


  function invalidChar(c) {
    if (c === undefined) {
      return syntaxError("json-hand: invalid end of input at ".concat(line, ":").concat(column));
    }

    return syntaxError("json-hand: invalid character '".concat(formatChar(c), "' at ").concat(line, ":").concat(column));
  }

  function invalidEOF() {
    return syntaxError("json-hand: invalid end of input at ".concat(line, ":").concat(column));
  }

  function invalidToken() {
    if (token.type === 'eof') {
      return syntaxError("json-hand: invalid end of input at ".concat(line, ":").concat(column));
    }

    return syntaxError("json-hand: invalid token '".concat(token.value, "' at ").concat(line, ":").concat(column));
  }

  function formatChar(c) {
    var replacements = {
      '\'': '\\\'',
      '"': '\\"',
      '\\': '\\\\',
      '\b': '\\b',
      '\f': '\\f',
      '\n': '\\n',
      '\r': '\\r',
      '\t': '\\t',
      '\v': '\\v',
      '\0': '\\0',
      "\u2028": "\\u2028",
      "\u2029": "\\u2029"
    };

    if (replacements[c]) {
      return replacements[c];
    }

    if (c < ' ') {
      var hexString = c.charCodeAt(0).toString(16);
      return '\\x' + ('00' + hexString).substring(hexString.length);
    }

    return c;
  }

  function syntaxError(message) {
    var err = new SyntaxError(message);
    err.lineNumber = line;
    err.columnNumber = column;
    return err;
  }

  var util$1 = {
    isSpaceSeparator: function isSpaceSeparator(c) {
      switch (c) {
        case '\t':
        case '\v':
        case '\f':
        case ' ':
        case "\xA0":
        case "\uFEFF":
        case '\n':
        case '\r':
        case "\u2028":
        case "\u2029":
          return true;
      }

      return /[\u1680\u2000-\u200A\u202F\u205F\u3000]/.test(c);
    },
    isSpecial: function isSpecial(c) {
      switch (c) {
        case ',':
        case ';':
        case '[':
        case ']':
        case '{':
        case '}':
        case ':':
          return true;
      }

      return false;
    },
    isDigit: function isDigit(c) {
      return /[0-9]/.test(c);
    },
    isHexDigit: function isHexDigit(c) {
      return /[0-9A-Fa-f]/.test(c);
    }
  };

  var json_1 = json;

  function json(source) {
    source = String(source);
    return JSON.stringify(parse(source), null, 2);
  }

  var src = {
    format: format_1,
    parse: parse,
    json: json_1
  };
  var src_1 = src.format;
  var src_2 = src.parse;
  var src_3 = src.json;

  exports.default = src;
  exports.format = src_1;
  exports.parse = src_2;
  exports.json = src_3;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
